/*! ********************************************
 * SMFSequence.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.smf;
public class SMFSequence
{
// Instance -----
	private int _format = 0;
	private boolean _multiPacketSysEx = true;

	private SMFTrackEvent _track = null;  // Event List - "MTrk" Events
	public int division = 480;  // "MThd" divison
	// Arbitrary Loop
	private SMFTrackEvent _loopStart = null;
	private SMFTrackEvent _loopEnd = null;
// getter, setter ---
	public void setTrack(SMFTrackEvent trk)
	{
		clear();
		_track = trk;
	}
	public SMFTrackEvent getTrack(){ return _track; }
	public boolean getLoopable(){ return _loopStart != null; }
	public SMFTrackEvent getLoopStart(){ return _loopStart; }
	public SMFTrackEvent getLoopEnd(){ return _loopEnd; }
	public boolean setLoopPoint(){ return setLoopPoint(null, null); }
	public boolean setLoopPoint(SMFTrackEvent start, SMFTrackEvent end)
	{
		if (_track == null) return false;
		_loopStart = null;
		_loopEnd = null;
		if (start == null || end == null || start == end) return false;
		SMFTrackEvent _list = start;
		int _deltaTimes = 0;
		// check -->
		while (_list != null)
		{
			if (_list == end)
			{
				if (_deltaTimes == 0) return false;
				_loopStart = start;
				_loopEnd = end;
				return true;
			}
			_deltaTimes += _list.deltaTime;
			_list = _list.next;
		}
		return false;
	}
	public void clear()
	{
		_track = null;
		division = 480;
		_loopStart = null;
		_loopEnd = null;
	}
// Constructor ---
	public SMFSequence(){}
	public SMFSequence(byte[] bytes)
	{
		loadBytes(bytes, true);
	}
	public SMFSequence(byte[] bytes, boolean multiPacketSysEx)
	{
		loadBytes(bytes, multiPacketSysEx);
	}
	public boolean loadBytes(byte[] bytes, boolean multiPacketSysEx)
	{
	// Start -->
		// 'MThd' ---
		if (bytes.length <= 16) return false;
		_multiPacketSysEx = multiPacketSysEx;
		int _dataSize = bytes.length;
		ByteArray _ba = new ByteArray(bytes);
		//_ba.endian = ByteArray.BIG_ENDIAN;
		if (_ba.readInt32() != 0x4d546864) return false;
		_ba.position += 4;
		_format	= _ba.readUint16();
		_ba.position += 2;
		int _div = _ba.readUint16();
		if (_format >= 2 || (_div & 0x8000) != 0 ) return false;  // No Suport "Format2", "SMPTE"
		// 'MTrk' ---
		SMFTrackEvent _head = new SMFTrackEvent(0, 0x80, 0, 0, 0);  // dummy
		while (_ba.position < _dataSize)
		{
			int _chunkType = _ba.readInt32(),
			    _chunkDataSize = _ba.readInt32();
			switch (_chunkType)
			{
				default : _ba.position += _chunkDataSize; break;  // Unknown Chunk
				case 0x4d54726b : 
					if (_read_MTrk(_ba, _chunkDataSize, _head)) break;
					else return false;
			}
		}
	// SetUp -->
		clear();
		_track = _head.next;
		if (_format != 0)_removeUnecEvents(_track);
		division = _div;
	// <-- Finish
		return true;
	}
	private boolean _read_MTrk(ByteArray bytes, int chunkDataSize, SMFTrackEvent trk)
	{
		int _chunkEnd = bytes.position + chunkDataSize;
		
		int _eventType = 0, _channel = 0;
		int _skipTime = 0;
		SMFTrackEvent _event = null;
		int _sysExDataSize, _sysExPacketsSize, _sysExPackets, _sysExTotalSize;
		int _metaType, _metaDataSize;
		
		// Start -->
		while (bytes.position < _chunkEnd)
		{
		// Read DeltaTime ---
			int _deltaTime = bytes.readVarLeng();
			if (_skipTime > 0){ _deltaTime += _skipTime; _skipTime = 0; }
		// Read StatusByte ---
			int _val = bytes.readUint8();
			if ((_val & 0x80) != 0)
			{
				int _status = _val;
				if (_status < 0xF0) { _eventType = _status & 0xf0; _channel = _status & 0xf; }  // Channel Message
				else { _eventType = _status; _channel = 0; }  // SysEx, MetaEvent
			}
			else bytes.position--;  // Runnig Status
			
			switch (_eventType)
			{
				default : bytes.position = _chunkEnd; return false; // unknown
			// MIDI Channel Message ---
				case 0x80 : case 0x90 : case 0xA0 : case 0xB0 : case 0xE0 :
					_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, bytes.readInt8(), bytes.readInt8());
					trk = _insertTrackEvent(trk, _event);
					continue;
					
				case 0xC0 : case 0xD0 : 
					_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, bytes.readInt8(), 0);
					trk = _insertTrackEvent(trk, _event);
					continue;
					
			// System Exclusive ---
				case 0xF0 : 
					_event = new SMFTrackEvent(_deltaTime, 0xF0, 0, 0, 0);
					trk = _insertTrackEvent(trk, _event);
					// Multi Packet Message -->
					if (_multiPacketSysEx)
					{
						// Read SysExData
						_sysExDataSize = bytes.readVarLeng();
						_event.data = new byte[_sysExDataSize];
						bytes.readBytes(_event.data, 0, _sysExDataSize);
						_event.dataSize = _sysExDataSize;
					}
					// Single Packet Message -->
					else
					{
						int _sysExStart = bytes.position;
						_skipTime = 0;
						// calc DataSize -->
						_sysExPacketsSize = 0;
						_sysExPackets = 0; 
						_sysExTotalSize = 0;
						while(true)
						{
							_sysExPacketsSize = bytes.readVarLeng();
							_sysExTotalSize += _sysExPacketsSize;
							bytes.position += _sysExPacketsSize;  // n Byte seaq
							_sysExPackets++;
							if ((bytes.bytes[bytes.position - 1] & 0xff) == 0xF7) break;  // EOX
							_skipTime += bytes.readVarLeng();
							bytes.position++; // status F7
						}
						//int _sysExEnd = bytes.position;
						// read Data -->
						_event.data = new byte[_sysExTotalSize];
						_event.dataSize = _sysExTotalSize;
						int p = 0;
						bytes.position = _sysExStart;
						while (true)
						{
							_sysExPacketsSize = bytes.readVarLeng();
							bytes.readBytes(_event.data, p, _sysExPacketsSize);
							if (--_sysExPackets > 0) break;
							p += _sysExPacketsSize;
							while ((bytes.bytes[bytes.position++] & 0x80) != 0);  // skip VarLengData
							bytes.position++; // status F7
						}
						//bytes.position = _sysExEnd;
					}
					continue;

				case 0xF7 : 
					if (_multiPacketSysEx)
					{
						_event = new SMFTrackEvent(_deltaTime, 0xF7, 0, 0, 0);
						trk = _insertTrackEvent(trk, _event);
						_sysExDataSize = bytes.readVarLeng();
						_event.data = new byte[_sysExDataSize];
						bytes.readBytes(_event.data, 0, _sysExDataSize);
						_event.dataSize = _sysExDataSize;
					}
					// Non-standard error
					else
					{
						bytes.position = _chunkEnd;
						return false;
					}
					continue;

			// Meta Event ---
				case 0xFF :
					_metaType = bytes.readUint8();
					_event = new SMFTrackEvent(_deltaTime, 0xFF00 | _metaType, 0, 0, 0);
					trk = _insertTrackEvent(trk, _event);
					// Read MetaData
					_metaDataSize = bytes.readVarLeng();
					if (_metaDataSize > 0)
					{
						_event.data = new byte[_metaDataSize];
						_event.dataSize = _metaDataSize;
						bytes.readBytes(_event.data, 0, _metaDataSize);
					}
					switch (_metaType)
					{
						case 0x2F: bytes.position = _chunkEnd; return true; // End of Track
						case 0x00: _event.value0 = (_event.data[0]&0xff)<<8 | (_event.data[1]&0xff); break; // Sequence Number
						case 0x20: _event.value0 = _event.data[0] & 0xff; break; // Channel Prefix
						case 0x51: _event.value0 = (_event.data[0]&0xff)<<16 | (_event.data[1]&0xff)<<8 | (_event.data[2]&0xff); break; // Set Tempo
					}
					continue;
			}
		}
		// <-- Finish
		bytes.position = _chunkEnd;
		return true;
	}
	private SMFTrackEvent _insertTrackEvent(SMFTrackEvent crnt, SMFTrackEvent ev)
	{
		if (_format == 0) return crnt.next = ev;
		
		SMFTrackEvent _next = crnt.next;
		while (_next != null)
		{
			if (ev.deltaTime >= _next.deltaTime) ev.deltaTime -= _next.deltaTime;
			else
			{
				_next.deltaTime -= ev.deltaTime;
				ev.next = _next;
				return crnt.next = ev;
			}
			crnt = _next;
			_next = _next.next;
		}
		return crnt.next = ev;
	}
	private void _removeUnecEvents(SMFTrackEvent trk)
	{
		if (trk == null)return;
		while (trk.next != null)
		{
			if (trk.next.type == 0xFF2F)
			{
				SMFTrackEvent _eotEv = trk.next;
				if (_eotEv.next == null) return;  // last EOT
				_eotEv.next.deltaTime += _eotEv.deltaTime;
				trk.next = _eotEv.next;
			}
			trk = trk.next;
		}
	}
	
	private static final class ByteArray
	{
		byte[] bytes = null;
		int position = 0;
	// Method
		void readBytes(byte[] bytes, int pos, int len)
		{
			int _end = pos + len;
			while (pos < _end)bytes[pos++] = this.bytes[position++];
		}

		byte readByte(){ return bytes[position++]; }
		int readInt8(){ return (int)bytes[position++]; }
		int readUint8(){ return (bytes[position++]&0xff); }
		int readInt16(){ return bytes[position++] << 8 | (bytes[position++]&0xff); }
		int readUint16(){ return (bytes[position++]&0xff) << 8 | (bytes[position++]&0xff); }
		int readInt32(){ return (bytes[position++]&0xff) << 24 | (bytes[position++]&0xff) << 16 | (bytes[position++]&0xff) << 8 | (bytes[position++]&0xff); }
		int readVarLeng()
		{
			int _rv = 0, _v = 0;
			do {
				_rv = bytes[position++] & 0xff;
				_v = _v << 7 | _rv & 0x7f;
			} while ((_rv & 0x80) != 0);
			return _v;
		}
	// Constructor
		public ByteArray(byte[] bytes)
		{
			this.bytes = bytes;
			position = 0;
		}
	}
}